{***********************************************************************
Unit ExtImage.pas.PAS v1.4 08/2001
    (c) 1999 by Andreas Moser, AMoser@t-online.de

    Delphi version : Delphi 3

    ExtImage.pas is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    which are not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

********************************************************************************}
unit ExtImage;

{$define USE_FULL_CAPABILITY}

interface
uses controls,forms,classes,windows,graphics, ExtCtrls, stdctrls, gfx_files, gfx_colors, gfx_effects,gfx_basedef,gfx_rotresize,gfx_colorpalette,Messages,sysutils;

type TZoomEvent=procedure (Sender : TObject ; Zoom:Double;RangeWidth,RangeHeight:Integer) of object;
     TScrollBarChangedEvent=procedure (Sender: TObject; Min,Max,Position:Integer) of Object;


TExtImage = class(TPanel)
protected
private
    OldWindowProc: TWndMethod;

    //image vars
    fImage:TImage;
    fImageLoaded:Boolean;
    fOnImageMouseMove:TMouseMoveEvent;
    fOnImageMouseUp:TMouseEvent;
    fOnImageMouseDown:TMouseEvent;
    fOnImageClick:TNotifyEvent;
    fOnStartSelect:TNotifyEvent;
    fOnAfterRedraw:TNotifyEvent;
    fOnAfterSelect:TNotifyEvent;

    //bitmap vars
    fBitmap:TBitmap;
    fOnBitmapChanged:TNotifyEvent;
    fZoomBitmap:TBitmap;
    fBufferBitmap:TBitmap;



    //zoom vars
    fZoom:Double;
    fZFactor:Double;
    fZoomEvent:TZoomEvent;
    fZoomEventOccured:Boolean;

    //Point vars
    fCursorX,fCursorY:Integer;
    fRelativeX,fRelativeY:Integer;
    fOriginalX,fOriginalY:Integer;
    fOrigin,fOMovePt,fDestPt,fZmOrigin,fZmDestPt:TPoint;
    fX,fY:Integer;
    fVertPixel:Integer;
    fHorzPixel:Integer;
    fPixelWidth:Integer;
    fMoving:Boolean;
    fMovePt:TPoint;
    fShift:TShiftState;

    // scrollbar vars
    fVPos,fHPos:Integer;
    fHBar,fVBar:TScrollBar;
    fVBarWidth:Integer;
    fHBarHeight:Integer;
    fOnHScrollBarChanged:TScrollBarChangedEvent;
    fOnVScrollBarChanged:TScrollBarChangedEvent;
    fScrollEventOccured:Boolean;



    //misc vars
    fConverterParam1:Integer;
    fConverterParam2:Integer;
    fPixelColor:TColor;
    fPixelRGB:TRGBTriple;
    fRGBLabel:TLabel;

    fAutoFit:Boolean;
    fDoNothing:Boolean;
    fShowXYCaption:Boolean;

    fVisibleHorz:Integer;
    fVisibleVert:Integer;

    fSizeMode: TSizeMode;
    fApplyMonitorGamma:Boolean;
    fMonitorGamma:Double;
    fGammaArray:Array [0..255] of Byte;
    fDrawing:Boolean;
    fFileName:String;

    //ruler vars
    fShowRuler:Boolean;
    fShowRulerDots:Boolean;
    fRuler_X,fRuler_Y:Integer;
    _rdelta_x,_rdelta_y:Integer;
    _rulerdiff:Integer;
    _ruler_x_bitmap:TBitmap;
    _ruler_y_bitmap:TBitmap;
    fRulerFirst:Boolean;

    //selectframe vars
    fSelectMode:Boolean;
    fSelected:Boolean;
    fSelMove:Boolean;
    fdrawSelection:Boolean;
    fblink_active:Boolean;
    _select_top_bitmap:TBitmap;
    _select_bottom_bitmap:TBitmap;
    _select_left_bitmap:TBitmap;
    _select_right_bitmap:TBitmap;
    sel_point_tl,sel_point_br:TPoint;


    fTimer:TTimer;


    function RasterX(Val:Integer):Integer;
    function RasterY(Val:Integer):Integer;
//    procedure CheckRasterBoundsX(var l1,l2:Integer;Min,Max:Integer);
    procedure CheckRasterBoundsY(var l1,l2:Integer;Min,Max:Integer);
    procedure _SetZoom(Value:Double);
    procedure _SetRuler(Value:Boolean);
    procedure _SetRulerDots(Value:Boolean);
    procedure _SetMonitorGamma(Value:Boolean);
    procedure _SetMonitorGammaVal(Value:Double);
    procedure _vScrollChange(Sender:TObject);
    procedure _hScrollChange(Sender:TObject);
    procedure _ImageMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure _ImageMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState;X, Y: Integer);
    procedure _ImageMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState;X, Y: Integer);
    procedure _ImageMouseClick(Sender: TObject);
    procedure _OnTimer(Sender: TObject);
    procedure _SetSelectMode(Value:Boolean);
    procedure _ClearSelection(Factor:Real);
    procedure _MakeSelection(Factor:Real;Mode:TPenMode);
    procedure _DrawSelection(TopLeft, BottomRight: TPoint);
    procedure _BlinkSelection;
    procedure _BitmapChanged(Sender:TObject);
    procedure _AutoFit;
    procedure _SetConstraint;
    procedure  NewWindowProc(var Message: TMessage);
    procedure  CM_ImageMouseLeave;
    procedure _ClearRulerDots;
    procedure _DrawRulerDots(x,y:Integer);
    procedure ScaleRect(var rRect:TRect);
public
    procedure   Resize;override;
    constructor Create(Aowner:TComponent);override;
    destructor  Destroy;override;
    procedure   Paint;override;
    function    CanResize(var NewWidth, NewHeight: Integer): Boolean;override;
    procedure   CreateWnd;override;
    procedure   Loaded;override;
    procedure   LoadImage(FileName:String);
    procedure   SaveImage(FileName:String);
    procedure   ReDraw;
    procedure   UpdatePalette;
    procedure   GetDimensions(AOrigin,ADestPt:TPoint;var rWidth,rHeight:Integer;var rRect:TRect);
    procedure   UpdateSelection;
    {$ifdef USE_FULL_CAPABILITY}
    procedure   ReduceColors(ColorReduction:TColorReduction;DitherMode:TDitherMode;ReductionBits:Integer;CustomPalette:HPalette);
    {$endif}
    property    Filename:String read fFileName;
    property    CursorXPos:Integer read fCursorX;
    property    CursorYPos:Integer read fCursorY;
    property    RelativeXPos:Integer read fRelativeX;
    property    RelativeYPos:Integer read fRelativeY;
    property    OriginalXPos:Integer read fOriginalX;
    property    OriginalYPos:Integer read fOriginalY;
    Property    Origin:TPoint Read fOrigin;
    Property    DestPt:TPoint Read fDestPt;
    Property    MovePt:TPoint Read fOMovePt;
    Property    ZMOrigin:TPoint Read fzmOrigin;
    Property    ZMDestPt:TPoint Read fzmDestPt;
    property    HSBarPos:Integer read fHPos;
    property    VSBarPos:Integer read fVPos;
    property    Bitmap:TBitmap read FBitmap;
    property    Image:TImage read FImage;
    property    Shift:TShiftState read fShift;
    property    Selected:Boolean read fSelected;
    property    ConverterParam1:Integer read fConverterParam1 write fConverterParam1;
    property    ConverterParam2:Integer read fConverterParam2 write fConverterParam2;
    property    HBarHeight:Integer read fHBarHeight;
    property    VBarWidth:Integer read  fVBarWidth;
    property    SizeMode:TSizeMode read fSizeMode write fSizeMode;
    property    VisibleVRange:Integer read fVisibleVert;
    property    VisibleHRange:Integer read fVisibleHorz;
    property    PixelColor:TColor read fPixelColor;
    property    PixelRGB:TRGBTriple read fPixelRGB;
published
    property RGBLabel:TLabel read fRGBLabel write fRGBLabel;
    property SelectMode:Boolean read fSelectMode write _SetSelectMode;
    property ShowRuler:Boolean read fShowRuler write _SetRuler;
    property ShowRulerDots:Boolean read fShowRulerDots write _SetRulerDots;
    property ShowXYCaption:Boolean read fShowXYCaption write fShowXYCaption;
    property MonitorGamma:Double read fMonitorGamma write _SetMonitorGammaVal;
    property ApplyMonitorGamma:Boolean read fApplyMonitorGamma write _SetMonitorGamma;
    property Zoom:Double read fZoom write _SetZoom;
    property OnImageMouseMove:TMouseMoveEvent read fOnImageMouseMove write fOnImageMouseMove;
    property OnImageMouseUp:TMouseEvent read fOnImageMouseUp write fOnImageMouseUp;
    property OnImageMouseDown:TMouseEvent read fOnImageMouseDown write fOnImageMouseDown;
    property OnImageClick:TNotifyEvent read fOnImageClick write fOnImageClick;
    property OnBitmapChanged:TNotifyEvent read fOnBitmapChanged write fOnBitmapChanged;
    property AfterZoom:TZoomEvent read fZoomEvent write fZoomEvent;
    property OnStartSelectMode:TNotifyEvent read fOnStartSelect write fOnStartSelect;
    property OnAfterSelect:TNotifyEvent read fOnAfterSelect write fOnAfterSelect;
    property OnAfterRedraw:TNotifyEvent read fOnAfterRedraw write fOnAfterRedraw;
    property OnHScrollBarChanged:TScrollBarChangedEvent read fOnHScrollBarChanged write fOnHScrollBarChanged;
    property OnVScrollBarChanged:TScrollBarChangedEvent read fOnVScrollBarChanged write fOnVScrollBarChanged;
    property AutoFit:Boolean read fAutoFit write fAutoFit;
end;


procedure Register;
var fff:Boolean;
const buffer_width=1;
implementation

constructor TExtImage.Create(Aowner:TComponent);
//---------------------------------------------------------------
//  Purpose: Constructor
//---------------------------------------------------------------
begin
    inherited Create(AOwner);
    fHBarHeight:=0;
    fVBarWidth:=0;
    Width:=200;
    Height:=100;
    fImageLoaded:=False;
    fDrawing:=False;

    fImage:=TImage.Create(self);
    fImage.Parent:=self;
//    fImage.Width:=Width;
//    fImage.Height:=Height;
    fImage.Visible:=false;
    fSizeMode:=smOriginal;

    fDoNothing:=False;

    fBitmap:=TBitmap.Create;
    fBitmap.OnChange:=_BitmapChanged;
    fZoomBitmap:=TBitmap.Create;

    fBitmap.PixelFormat:=pf24Bit;
    fZoomBitmap.PixelFormat:=pf24Bit;

    fBitmap.IgnorePalette:=True;
    fZoomBitmap.IgnorePalette:=True;
    fImage.Picture.Bitmap.IgnorePalette:=True;

    if (GetDeviceCaps(GetDC(0),BITSPIXEL)=8) then
    begin
      fImage.Picture.Bitmap.IgnorePalette:=False;
      fImage.Picture.Bitmap.PixelFormat:=pf8Bit;
    end;

    fZoom:=100;
    fZFactor:=1;
    fVpos:=0;
    fHPos:=0;

    fAutoFit:=False;

    fHBar:=TScrollBar.Create(Self);
    fHBar.Parent:=self;
    fHBar.Kind:=sbHorizontal;
    fHBar.Visible:=false;
    fVBar:=TScrollBar.Create(Self);
    fVBar.Parent:=self;
    fVBar.Kind:=sbVertical;
    fVBar.Visible:=false;

    fRuler_Y:=0;
    fRuler_X:=0;
    fShowRuler:=False;
    _rdelta_x:=0;
    _rdelta_y:=0;
    _rulerdiff:=5;
    _ruler_x_bitmap:=TBitmap.Create;
    _ruler_y_bitmap:=TBitmap.Create;
    fRulerFirst:=True;

    _select_top_bitmap:=TBitmap.Create;
    _select_bottom_bitmap:=TBitmap.Create;
    _select_left_bitmap:=TBitmap.Create;
    _select_right_bitmap:=TBitmap.Create;
    fblink_active:=false;

    fBufferBitmap:=TBitmap.Create;

    fHBar.OnChange:=_hScrollChange;
    fVBar.OnChange:=_vScrollChange;

    fImage.OnClick:=_ImageMouseClick;
    fImage.OnMouseMove:=_ImageMouseMove;
    fImage.OnMouseUp:=_ImageMouseUp;
    fImage.OnMouseDown:=_ImageMouseDown;

    fImage.Width:=Width;
    fImage.Height:=Height;

    fCursorX:=0;
    fCursorY:=0;

    fVertPixel:=0;
    fHorzPixel:=0;

    fX:=0;
    fY:=0;

    fdrawSelection:=False;
    fSelectMode:=False;
    fSelected:=False;

    fZoomEventOccured:=False;
    fScrollEventOccured:=False;

    fApplyMonitorGamma:=false;
    fMonitorGamma:=1.0;
    OldWindowProc := fImage.WindowProc; //store its default Window method
    fImage.WindowProc := NewWindowProc; //change its Window method to this one

    fTimer:=TTimer.Create(self);
    fTimer.Interval:=150;
    fTimer.Enabled:=False;
    fTimer.OnTimer:= _OnTimer;

    fSelMove:=False;
    fMoving:=False;

    fFileName:='';
    fShowXYCaption:=False;


end;

procedure TExtImage.NewWindowProc(var Message: TMessage);
{the new WindowProc of the Win97Control--
 first performs all default functionality and then handles each
 message for Win97 functionality}
begin
  OldWindowProc(Message);   //perform default method
  if (self <> nil) and not (csDesigning in ComponentState) then
    try
      Case Message.Msg of
         CM_MouseLeave: CM_ImageMouseLeave;
       end;
    except
    end;
end;

procedure TExtImage.CM_ImageMouseLeave;
begin
   _ClearRulerDots;
   _DrawRulerDots(-1,-1);

end;

destructor TExtImage.Destroy;
//---------------------------------------------------------------
//  Purpose: Destructor
//---------------------------------------------------------------
begin
    fBitmap.free;
    fZoomBitmap.free;
    _ruler_x_bitmap.Free;
    _ruler_y_bitmap.Free;
    _select_top_bitmap.Free;
    _select_bottom_bitmap.Free;
    _select_left_bitmap.Free;
    _select_right_bitmap.Free;
    fBufferBitmap.Free;
    inherited;
end;

procedure TExtImage.Paint;
var i:Integer;
begin
 if (csLoading in ComponentState) or (csReading in ComponentState)
 or (csDesigning in ComponentState) then
 begin
   inherited Paint;
   exit;
 end
 else
  if fImageLoaded and fShowRuler then
  begin
    Canvas.Pen.Color:=clWhite;

    Canvas.MoveTo(fruler_X,4);
    Canvas.LineTo(Width-fruler_X,4);

    Canvas.MoveTo(fruler_X,fimage.Height+2*fruler_y-5);
    Canvas.LineTo(Width-fruler_X,fimage.Height+2*fruler_y-5);

    Canvas.MoveTo(4,fruler_y);
    Canvas.LineTo(4,Height-fruler_y);

    Canvas.MoveTo(fimage.Width+2*fruler_x-5,fruler_y);
    Canvas.LineTo(fimage.Width+2*fruler_x-5,Height-fruler_y);

    Canvas.Pen.Color:=clGray;

    Canvas.MoveTo(fruler_X,5);
    Canvas.LineTo(Width-fruler_X,5);

    Canvas.MoveTo(fruler_X,fimage.Height+2*fruler_y-4);
    Canvas.LineTo(Width-fruler_X,fimage.Height+2*fruler_y-4);

    Canvas.MoveTo(5,fruler_y);
    Canvas.LineTo(5,Height-fruler_y);

    Canvas.MoveTo(fimage.Width+2*fruler_x-4,fruler_y);
    Canvas.LineTo(fimage.Width+2*fruler_x-4,Height-fruler_y);


    Canvas.Pen.Color:=clBlack;
    for i:=0 to fImage.Width do
    begin
      if (i mod 5 = 0) and (i mod 10 <> 0)  then
      begin
        Canvas.MoveTo(i+fruler_x,fruler_y-1);
        Canvas.LineTo(i+fruler_x,fruler_y-4);
        Canvas.MoveTo(i+fruler_x,fruler_y+fImage.Height+1);
        Canvas.LineTo(i+fruler_x,fruler_y+fImage.Height+4);
      end else
      if i mod 10 = 0 then
      begin
        Canvas.MoveTo(i+fruler_x,fruler_y-8);
        Canvas.LineTo(i+fruler_x,fruler_y-1);
        Canvas.MoveTo(i+fruler_x,fruler_y+fImage.Height+1);
        Canvas.LineTo(i+fruler_x,fruler_y+fImage.Height+8);
      end;
    end;

    for i:=0 to fImage.height do
    begin
      if (i mod 5 = 0) and (i mod 10 <> 0)  then
      begin
        Canvas.MoveTo(fruler_x-1,i+fruler_y);
        Canvas.LineTo(fruler_x-4,i+fruler_y);
        Canvas.MoveTo(fruler_x+fImage.Width+1,i+fruler_y);
        Canvas.LineTo(fruler_x+fImage.Width+4,i+fruler_y);
      end else
      if i mod 10 = 0 then
      begin
        Canvas.MoveTo(fruler_x-1,i+fruler_y);
        Canvas.LineTo(fruler_x-8,i+fruler_y);
        Canvas.MoveTo(fruler_x+fImage.Width+1,i+fruler_y);
        Canvas.LineTo(fruler_x+fImage.Width+8,i+fruler_y);
      end;
    end;
  end else inherited;
end;

procedure TExtImage.ReDraw;
//---------------------------------------------------------------
//  Purpose: Redraw the image
//---------------------------------------------------------------
var zW,zH,czW,czH,zx,zy:Integer;
    wid,hei:Integer;
    rec:TRect;
    pRow:pRGBarray;
    pal:HPalette;
    hch,wch:boolean;
    wcv:Integer;
    pw:Integer;

// Gamma correction if necessary
procedure Gamma;
var
    row,col:Integer;
begin
      if fApplyMonitorGamma then
      begin
        for row:=0 to fZoomBitmap.Height-1 do
        begin
          pRow:=fZoomBitmap.ScanLine[row];
          for col:=0 to fZoomBitmap.Width-1 do
          begin
            pRow[col].rgbtBlue:=fGammaArray[pRow[col].rgbtBlue];
            pRow[col].rgbtRed:=fGammaArray[pRow[col].rgbtRed];
            pRow[col].rgbtGreen:=fGammaArray[pRow[col].rgbtGreen];
          end;
        end;
      end;
      if GetDeviceCaps(GetDC(0),BITSPIXEL)=8 then
      begin
        pal:=CreateOptimizedPaletteforSingleBitmap(fBitmap,6);
        ApplyPaletteToGraphic(pal,fImage.Picture.Bitmap,true,pf8bit);
      end;
end;

begin
 if (csLoading in ComponentState) or (csReading in ComponentState)
 or (csDesigning in ComponentState) then
 begin
   fHBar.Width:=Width;
   fVBar.Height:=Height;
   fImage.Width:=Width;
   fImage.Height:=Height;
   exit;
 end
 else
 begin
   fTimer.Enabled:=False;
   fVisibleHorz:=0;
   fVisibleVert:=0;

   // Get actual clientsize values
   zw:=ClientWidth-(2*fRuler_X);
   zh:=ClientHeight-(2*fRuler_Y);



   // if clientarea is to small, exit
   if (zw<5) or (zh<5) then exit;

   pw:= Round(fZFactor);
   if pw<1 then pw:=1;
   //set defaultvalues for canvas etc.
   fBitmap.Canvas.CopyMode:=cmSrcCopy;
   fimage.picture.Bitmap.Canvas.CopyMode:=cmSrcCopy;
   fBitmap.Canvas.Pen.style:=psSolid;
   fimage.picture.Bitmap.Canvas.Pen.style:=fBitmap.Canvas.Pen.style;
   fBitmap.Canvas.Brush.style:=bsClear;
   fimage.picture.Bitmap.Canvas.Brush.style:=fBitmap.Canvas.Brush.style;
   fZoomBitmap.Canvas.CopyMode:=cmSrcCopy;
   fZoomBitmap.Canvas.Pen.style:=psSolid;
   fZoomBitmap.Canvas.Brush.style:=bsClear;
   fBufferBitmap.Canvas.CopyMode:=cmSrcCopy;


   //prevent events like resize to do something
   fDoNothing:=True;

   //initalize switches and values
   hch:=False;
   wch:=False;
   wcv:=0;

   if zw>=(fBitmap.Width * fzFactor) then zw:= Round(fBitmap.Width * fzFactor);
   if zh>=(fBitmap.Height * fzFactor) then zh:= Round(fBitmap.Height * fzFactor);

   // if the width of an image is higher than its height proceed this
   if fBitmap.Width>fBitmap.height then
   begin
     if (fBitmap.Width * fzFactor)>zw then
     begin
       fHBar.Visible:=True;
       zh:=zh-16;
       hch:=true;
     end else fHBar.Visible:=False;
     if (fBitmap.Height * fzFactor)>zh then
     begin
        fVBar.Visible:=True;
        zw:=zw-16;
        if fBitmap.Width * fzFactor >zw then
        begin
          fHBar.Visible:=True;
          if not hch then zh:=zh-16;
        end;
     end else fVBar.Visible:=False;
   end
   else
   // otherwise proceed this stuff
   begin
     if (fBitmap.Height * fzFactor)>zh then
     begin
        fVBar.Visible:=True;
        zw:=zw-16;
        wcv:=16;
        wch:=True;
     end else
     begin
       fVBar.Visible:=False;
       fVPos:=0;
     end;

     if ((fBitmap.Width * fzFactor)> zw+1) then
     begin
       fHBar.Visible:=True;
       zh:=zh-16;
       if (fBitmap.Height * fzFactor)>zh+1 then
       begin
         fVBar.Visible:=True;
         if not wch then zw:=zw-wcv;
       end;
     end else
     begin
       fHBar.Visible:=False;
       fHPos:=0;
     end;
   end;
   //the scrollbar-init seem to be finished


   //drawing starts here
   fImage.Left:=fRuler_x;
   fImage.top:=fRuler_y;


   fimage.Width:=Abs(zw);
   fimage.Height:=Abs(zh);                 //...until here

   if fZoom >=100 then
   begin

     fimage.picture.Bitmap.Width:=Abs(((zw div pw) * pw)  {+pw} );   //Set visible range ...
     fimage.picture.Bitmap.Height:=Abs(((zh div pw) * pw) { +pw });
     fBufferBitmap.Width:=fimage.picture.Bitmap.Width;           //Set visible range (buffer bitmap)
     fBufferBitmap.Height:=fimage.picture.Bitmap.Height;

     if (zw >= fBitmap.Width * fZFactor) then pw:=1;
     if (zh >= fBitmap.Height * fZFactor) then pw:=1;

     czW:=(zw div pw){+1};
     czH:=(zh div pw){+1};

     if czW> FBitmap.Width then czW:=  FBitmap.Width;
     if czh> FBitmap.Height then czh:=  FBitmap.Height;


     fZoomBitmap.Width:=czW;             //Zoombmp is the buffer for drawing
     fZoomBitmap.Height:=czH;            // setting dimensions and copy source bmp to zoombmp

    // check upper bounds
    if fHPos+czw >= fBitmap.Width then
    zx:=fBitmap.Width-czW
    else zx:=fHPos;
    if fVPos+czH >= fBitmap.Height then
    zy:=fBitmap.Height-czH
    else  zy:=fVPos;
    if zx<0 then zx:=0;
    if zy<0 then zY:=0;

    fZoomBitmap.Canvas.CopyRect(Rect(0,0,czW,czH),fBitmap.Canvas,Rect(zx,zy,zx+czW,zy+czH));
    fX:=zx;
    fY:=zy;

    //finally, put the buffer to the visible area
    Gamma;
    fImage.Picture.Bitmap.Canvas.StretchDraw(
    Rect(0,0,fimage.picture.Bitmap.Width,fimage.picture.Bitmap.Height),fZoomBitmap);
    fBufferBitmap.Canvas.StretchDraw(Rect(0,0,fBufferBitmap.Width,fBufferBitmap.Height),fZoomBitmap);
   end else
   begin

      fimage.picture.Bitmap.Width:=Abs(zw);   //Set visible range ...
      fimage.picture.Bitmap.Height:=Abs(zh);
      fBufferBitmap.Width:=fimage.picture.Bitmap.Width;           //Set visible range (buffer bitmap)
      fBufferBitmap.Height:=fimage.picture.Bitmap.Height;
      czW:=Round(fBitmap.Width*fzFactor);
      czH:=Round(fBitmap.Height*fzFactor);

      if fHPos>fBitmap.Width - czW then fHPos:=0;
      if fVPos>fBitmap.Height - czH then fVPos:=0;

      pw:=1;
      fZoomBitmap.Width:=czW;
      fZoomBitmap.Height:=czH;
      if czW<ClientWidth then begin
        fImage.Picture.Bitmap.Width:=czW;
      end;
      if czH<Height then begin
        fImage.Picture.Bitmap.Height:=czH;
      end;
      fZoomBitmap.Canvas.StretchDraw(Rect(0,0,czW,czH),FBitmap);
      fX:=0;
      fY:=0;
      Gamma;
      fImage.Picture.Bitmap.Canvas.CopyRect(Rect(0,0,fImage.Picture.Bitmap.Width,fImage.Picture.Bitmap.Height),
      fZoomBitmap.Canvas,Rect(fHPos,fVPos,fHPos+fImage.Picture.Bitmap.Width,fVPos+fImage.Picture.Bitmap.Height));
      fBufferBitmap.Canvas.CopyRect(Rect(0,0,fBufferBitmap.Width,fBufferBitmap.Height),
      fZoomBitmap.Canvas,Rect(fHPos,fVPos,fHPos+fBufferBitmap.Width,fVPos+fBufferBitmap.Height))

   end;
   fVertPixel:=czH;
   fHorzPixel:=czW;
   fPixelWidth:=pw;

   //drawing finished

   fVisibleHorz:=fImage.Width;
   fVisibleVert:=fImage.height;


   // set the values of the scrollbars if necessary
   if fVBar.Visible then
   begin
      fVBAr.Left:=fImage.Width+(2*fruler_x);
      fVBAr.Width:=16;
      fVBAr.Height:=fImage.Height;
      fVBAr.Top:=fruler_y;
      fVBarWidth:=fVBar.Width;
      if (fZoom >=100) and (fBitmap.Height-(fimage.Height div pw)>0) then
      fVBar.Max:=(fBitmap.Height-(fimage.Height div pw))
      else
      fVBar.Max:=Round(fBitmap.Height* fzFactor)-fImage.Height;
      fVBar.Min:=0;
      fVBar.LargeChange:=fVBar.Max div 3;
      fVBar.SmallChange:=fVBar.Max div 10;
      Inc(fVisibleHorz,16);
   end else begin
     fVBarWidth:=0;
   end;

   if fHBar.Visible then
   begin
      fHBAr.Top:=fImage.Height+(2*fruler_y);
      fHBAr.Height:=16;
      fHBAr.Width:=fImage.Width;
      fHBar.Left:=fruler_x;
      fHBarHeight:=fhBar.Height;
      if (fZoom >=100) and (fBitmap.Width-(fimage.Width div  pw)>0) then
      begin

        fHBar.Max:=fBitmap.Width-(fimage.Width div  pw)
      end
      else
      fHBar.Max:=Round(fBitmap.Width* fzFactor)-fImage.Width;
      fHBar.Min:=0;
      fHBar.LargeChange:=fHBar.Max div 3;
      fHBar.SmallChange:=fHBar.Max div 10;
      Inc(fVisibleVert,16);

   end else begin
     fHBarHeight:=0;
   end;

   // events dont need to be ignored by now


  fzmOrigin.X:=Round((fOrigin.X-fHPos) * fZFactor);
  fzmOrigin.Y:=Round((fOrigin.Y-fVPos) * fZFactor);
  fzmDestPt.X:=Round((fDestPt.X-fHPos) * fZFactor);
  fzmDestPt.Y:=Round((fDestPt.Y-fVPos) * fZFactor);
  if fSelected then
  begin
   GetDimensions(fzmOrigin,fzmDestPt,wid,hei,rec);
   fff:=True;
   if (
         (rec.Left <0) and (rec.Right <0) or
         (rec.Left >zw) and (rec.Right >zw)
      ) or
      (
         (rec.Top <0) and (rec.Bottom <0) or
         (rec.Top >zh) and (rec.Bottom >zh)
      )  then
   begin
     fTimer.Enabled:=False;
   end else
   begin
    _MakeSelection(fZFactor,pmNotXOr);
     if not fMoving and fblink_active then fTimer.Enabled:=True;
   end;
  end;
  fDoNothing:=False;

  if Assigned(FOnAfterRedraw) then FOnAfterRedraw(self);

 end;
end;

procedure TExtImage.Loaded;
//---------------------------------------------------------------
//  Purpose: Loaded event
//---------------------------------------------------------------
begin
    inherited Loaded;
    if (csDesigning in ComponentState) then
    begin
      fHBar.Visible:=false;
      fVBar.Visible:=false;
    end;
end;

procedure TExtImage.CreateWnd;
//---------------------------------------------------------------
//  Purpose: CreateWnd event
//---------------------------------------------------------------
begin
    inherited;
end;

procedure TExtImage.Resize;
//---------------------------------------------------------------
//  Purpose: Resize event
//---------------------------------------------------------------
begin
   if fDoNothing Then Exit;
   fHPos:=fHBar.Position;
   fVPos:=fVBar.Position;
   inherited Resize;
   ReDraw;
   //if not (csDestroying in ComponentState) then ReDraw;
end;

function TExtImage.CanResize(var NewWidth, NewHeight: Integer): Boolean;
begin
 result:=inherited CanResize(NewWidth,NewHeight);
end;

procedure TExtImage.LoadImage(FileName:String);
//---------------------------------------------------------------
//  Purpose: Load an image
//---------------------------------------------------------------
var pal:hPalette;
begin
  fImageLoaded:=False;
  fTimer.Enabled:=False;
  fblink_active:=False;
  fSelected:=False;
  Width:=200;
  Height:=100;
  LoadAnyImageToBMP(fBitmap,FileName,false,fConverterParam1,fConverterParam2);
  fFileName:=FileName;
  if Parent is TForm then TForm(Parent).Caption:=ExtractFileName(fFileName);
  fZoomBitmap.PixelFormat:=fBitmap.PixelFormat;
  fZoomBitmap.IgnorePalette:=fBitmap.IgnorePalette;
  if GetDeviceCaps(GetDC(0),BITSPIXEL)=8 then
  begin
    pal:= CreateOptimizedPaletteforSingleBitmap(FBitmap,6);
    fImage.Picture.Bitmap.Ignorepalette:=False;
    ApplyPaletteToGraphic(pal,fImage.Picture.Bitmap,true,pf8Bit);
    ApplyPaletteTODC(self.handle,pal);
  end else
  begin
   fImage.Picture.Bitmap.PixelFormat:=fBitmap.PixelFormat;
   fImage.Picture.Bitmap.IgnorePalette:=fBitmap.IgnorePalette;
  end;
  UpdatePalette;
   fImageLoaded:=True;
   fDoNothing:=True;
   case fSizeMode of
   smUseZoomValue:
              begin
                _AutoFit;
              end;
   smOriginal:begin
                fZoom:=100;
                _AutoFit;
              end;
   smFitBoth: begin
                 fZoom:=100;
                 repeat
                  fZoom:=fZoom-1;
                  if fZoom < 10 then break;
                 until (fBitmap.Height * (fZoom /100) < (Screen.Height-40))
                 and (fBitmap.Width * (fZoom /100) < (Screen.Width-10));
                 _AutoFit;
              end;
   smFitHeight:begin
                 if fBitmap.Height > (Screen.Height-40) then
                 begin
                   fZoom:=((Screen.Height-40) / fBitmap.Height)  * 100 ;
                 end else fZoom:=100;
                 _AutoFit;
               end;
   smFitWidth:begin
                 if fBitmap.Width > (Screen.Width-10) then
                 begin
                   fZoom:=((Screen.Width-10) / fBitmap.Width)  * 100 ;
                 end else fZoom:=100;
                 _AutoFit;
              end;
   end;
   fDoNothing:=False;
   Resize;
   ReDraw;
   fImage.Visible:=true;
   _SetConstraint;
end;

procedure TExtImage._AutoFit;
var
    d:Integer;
begin
  if (csDesigning in ComponentState) or not fImageLoaded then Exit;
  if fAutoFit then
  begin
    fDoNothing:=True;
    with Constraints do
    begin
     MinHeight:=0;
     MinWidth:=0;
     MaxHeight:=0;
     MaxWidth:=0;
    end;
    if Parent is TWinControl then
    begin
      if (fBitmap.Width * fzFactor)+(2*fRuler_X) <Screen.Width-((TWinControl(Parent).Width-(TWinControl(Parent).ClientWidth)))
      then TWinControl(Parent).ClientWidth:=Round(fBitmap.Width * fzFactor)+(2*fRuler_X)+1  else
      begin
        d:=0;
        if {Screen.Height -(TWinControl(Parent).Height-}(TWinControl(Parent).ClientWidth-(2*fRuler_x)) < Round(fBitmap.Width *fzFactor) then d:=16;
        TWinControl(Parent).ClientWidth:=Round(fBitmap.Width * fzFactor)+d+(2*fRuler_x)+1;
      end;
      if (fBitmap.Height * fzFactor)+(2*fRuler_Y) <Screen.Height-((TWinControl(Parent).Height-(TWinControl(Parent).ClientHeight)))
      then  TWinControl(Parent).ClientHeight:=Round(fBitmap.Height * fzFactor)+(2*fRuler_Y)+1  else
      begin
        d:=0;
        if {Screen.Height -(TWinControl(Parent).Height-}(TWinControl(Parent).ClientHeight-(2*fRuler_Y)) < Round(fBitmap.Height *fzFactor) then d:=16;
        TWinControl(Parent).ClientHeight:=Round(fBitmap.Height * fzFactor)+d+(2*fRuler_y)+1;
      end;
    end;
    fDoNothing:=False;
  end;
end;

procedure TExtImage._SetConstraint;
begin
   if (csDesigning in ComponentState) or not fImageLoaded then Exit;
   fDoNothing:=True;
   with Constraints do
   begin
     MinHeight:=5;
     MinWidth:=5;
     MaxHeight:=Round(fBitmap.Height * fzFactor)+(fRuler_Y*2)+fHBarHeight+1;
     MaxWidth:=Round(fBitmap.Width * fzFactor)+(fRuler_X*2)+fVBarWidth+1;
   end;
   fDoNothing:=False;
end;

procedure TExtImage.SaveImage(FileName:String);
//---------------------------------------------------------------
//  Purpose: save bmp to imagefile
//---------------------------------------------------------------
begin
  SaveBMPToAnyFile(fBitmap,FileName,fConverterParam1,fConverterParam2);
end;

procedure TExtImage._SetZoom(Value:Double);
//---------------------------------------------------------------
//  Purpose: Set the zoom value
//---------------------------------------------------------------
begin
  if Value<>fZoom then
  begin
    fHPos:=0;
    fVPos:=0;
    fx:=0;
    fy:=0;
    fZoomEventOccured:=True;
    if Value<=10 then Value:=10 else
    if Value>=2000 then Value:=2000;
    fZoom:=Value;
    fzFactor:=fZoom/100;
    if fShowRuler then begin
      _ClearRulerDots;
      _rdelta_x:=0;
      _rdelta_y:=0;
    end;
    _AutoFit;
    _SetConstraint;
    ReDraw;
    if Assigned(fZoomEvent) then fZoomEvent(self,fzoom,fBitmap.Width,fBitmap.Height);
  end;
end;

procedure TExtImage._SetRuler(Value:Boolean);
//---------------------------------------------------------------
//  Purpose: Set the ruler
//---------------------------------------------------------------
begin
  if Value<>fShowRuler then
  begin
    fShowRuler:=Value;
    if fShowRuler=True then
    begin
      fRuler_X:=24;
      fRuler_Y:=24;
    end else
    begin
      fRuler_X:=0;
      fRuler_Y:=0;
    end;
      _SetConstraint;
      _AutoFit;
    ReDraw;
  end;
end;

procedure TExtImage._SetRulerDots(Value:Boolean);
//---------------------------------------------------------------
//  Purpose: Set the ruler
//---------------------------------------------------------------
begin
  if Value<>fShowRulerDots then
  begin
    fShowRulerDots:=Value;
    ReDraw;
  end;
end;

procedure TExtImage._SetSelectMode(Value:Boolean);
//---------------------------------------------------------------
//  Purpose: Change selectMode
//---------------------------------------------------------------
begin
  if Value <> fSelectMode then fSelectMode:=Value;
  if (fSelectMode=False) and (fSelected=True) then _ClearSelection(fZFactor);
  if fSelectMode= True then if Assigned(fOnStartSelect) then fOnStartSelect(self);

end;

procedure TExtImage._hScrollChange(Sender:TObject);
//---------------------------------------------------------------
//  Purpose: HorzScrollBar change event
//---------------------------------------------------------------
begin
  fzmOrigin.X:=Round((fOrigin.X-fHPos) * fZFactor);
  fzmOrigin.Y:=Round((fOrigin.Y-fVPos) * fZFactor);
  fzmDestPt.X:=Round((fDestPt.X-fHPos) * fZFactor);
  fzmDestPt.Y:=Round((fDestPt.Y-fVPos) * fZFactor);
  fScrollEventOccured:=True;
  fHPos:=fHBar.Position;
  CM_ImageMouseLeave;
  ReDraw;
    if fSelected then begin

      fSelected:=true;
    end;
  if fHBAr.Visible then if Assigned(fOnHScrollBarChanged) then fOnHScrollBarChanged(self,fHBAr.Min,fHBar.Max,fHPos);
end;

procedure TExtImage._vScrollChange(Sender:TObject);
//---------------------------------------------------------------
//  Purpose: VertScrollBar change event
//---------------------------------------------------------------
begin

  fzmOrigin.X:=Round((fOrigin.X-fHPos) * fZFactor);
  fzmOrigin.Y:=Round((fOrigin.Y-fVPos) * fZFactor);
  fzmDestPt.X:=Round((fDestPt.X-fHPos) * fZFactor);
  fzmDestPt.Y:=Round((fDestPt.Y-fVPos) * fZFactor);
  fScrollEventOccured:=True;
  fVPos:=fVBar.Position;
  CM_ImageMouseLeave;
  ReDraw;
  if fSelected then begin

      fSelected:=true;
    end;

  if fVBAr.Visible then if Assigned(fOnVScrollBarChanged) then fOnVScrollBarChanged(self,fVBAr.Min,fVBar.Max,fVPos);
end;

procedure TExtImage._ImageMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
//---------------------------------------------------------------
//  Purpose: Mouse move event
//---------------------------------------------------------------
var dx,dy,hx,hy:Integer;
begin
  hx:=x;
  hy:=y;
  fMoving:=True;
  if (X>=fImage.Width) then x:=fImage.Width-1;
  if (X<=0) then x:=1;
  if (Y>=fImage.Height) then Y:=fImage.Height-1;
  if (Y<=0) then Y:=1;
  fShift:=Shift;

  fRelativeX:=X;
  fRelativeY:=Y;
  _ClearRulerDots;
  if fZoom>100 then dX:=fx+Round((X / fZFactor)) else dX
  :=Round(fx+X/fZFactor);
  if fZoom>100 then dY:=fy+Round((Y / fZFactor)) else dY
  :=Round(fy+Y/fZFactor);
  fCursorX:=dx;
  fCursorY:=dy;

  if (ssLeft in Shift) and fDrawSelection and fSelectMode then
  begin
     fSelected:=True;
     fSelMove:=True;
     try
     fOMovePt.X:=dx;
     fOMovePt.y:=dy;

       fMovePt := Point(X,Y);
       if hx > fImage.Width then if fHBAr.Position< fBitmap.Width-fHorzPixel then fHBar.Position:=fHBar.Position+4;
       if hx <= 0 then if fHBAr.Position>0 then fHBar.Position:=fHBar.Position-4;
       if hy > fImage.Height then if fVBAr.Position< fBitmap.Height-fVertPixel then fVBar.Position:=fVBar.Position+4;
       if hy <= 0 then if fVBAr.Position>0 then fVBar.Position:=fVBar.Position-4;
     finally
       fSelMove:=False;
     end;
  end;
  _DrawRulerDots(X,Y);
  fPixelColor:=fBitmap.Canvas.Pixels[dx,dy];
  fPixelRGB:=ColorToRGB(fPixelColor);
  if Assigned(FRGBLabel) then
  FRGBLabel.Caption:=IntToStr(fPixelRGB.rgbtRed)+';'+IntToStr(fPixelRGB.rgbtGreen)+';'+IntToStr(fPixelRGB.rgbtBlue);

  fMoving:=False;

  if Assigned(fOnImageMouseMove) then fOnImageMouseMove(self,shift,x,y);
end;

procedure TExtImage._ClearRulerDots;
begin
    if not fShowRuler then
    begin
     _DrawSelection(fOrigin, fOMovePt);
     exit;
    end;

    Canvas.Brush.Color:=Color;
    Canvas.Brush.Style:=bsSolid;
    //Clear
    Canvas.FillRect(Rect(_rdelta_x-4,1+_rulerdiff,_rdelta_x+4,9+_rulerdiff));
    Canvas.FillRect(Rect(_rdelta_x-4,2*fruler_y+fImage.Height-1-_rulerdiff,_rdelta_x+4,2*fruler_y+fImage.Height-9-_rulerdiff));
    Canvas.FillRect(Rect(1+_rulerdiff,_rdelta_y-4,9+_rulerdiff,_rdelta_y+4));
    Canvas.FillRect(Rect(2*fruler_x+fImage.Width-1-_rulerdiff,_rdelta_y-4,2*fruler_x+fImage.Width-9-_rulerdiff,_rdelta_y+4));
    if not fShowRuler and fRulerFirst then exit;

    if fSelected {and not fZoomEventOccured }then _DrawSelection(fOrigin, fOMovePt);

    if fShowRulerDots then
    begin
    fImage.Canvas.Pen.Style:=psDot;
    fImage.Canvas.Pen.Mode:=pmNot;
    fImage.Canvas.MoveTo(0,_rdelta_y-fRuler_y);
    fImage.Canvas.LineTo(fImage.Width,_rdelta_y-fRuler_y);
    fImage.Canvas.MoveTo(_rdelta_x-fRuler_x,0);
    fImage.Canvas.LineTo(_rdelta_x-fRuler_x,fImage.Height);
    end;

end;

procedure TExtImage._DrawRulerDots(x,y:Integer);
var tp,tp2:TPen;
begin

    if not fShowRuler then
    begin
      _DrawSelection(fOrigin, fOMovePt);
      exit;
    end;
    fRulerFirst:=False;
    tp:=Canvas.Pen;
    tp2:=fImage.Canvas.Pen;
    Canvas.Pen.Color:=clRed;
    Canvas.Pen.Width:=1;
    Canvas.Brush.Color:=clMaroon;


    _ruler_x_bitmap.Width:=fBufferBitmap.Width;
    _ruler_x_bitmap.Height:=1;
    _ruler_x_bitmap.PixelFormat :=fBufferBitmap.PixelFormat;

    _ruler_x_bitmap.Canvas.CopyRect(Rect(0,0,_ruler_x_bitmap.Width,1),fBufferBitmap.Canvas,
    Rect(0,y,_ruler_x_bitmap.Width,y+1));

    _ruler_y_bitmap.Height:=fBufferBitmap.Height;
    _ruler_y_bitmap.Width:=1;
    _ruler_y_bitmap.PixelFormat :=fBufferBitmap.PixelFormat;

    _ruler_y_bitmap.Canvas.CopyRect(Rect(0,0,1,_ruler_y_bitmap.Height),fBufferBitmap.Canvas,
    Rect(x,0,x+1,_ruler_y_bitmap.Height));


    //Top
    Canvas.Moveto(fRuler_x+x-3,_rulerdiff+2);
    Canvas.LineTo(fRuler_x+x+3,_rulerdiff+2);
    Canvas.LineTo(fRuler_x+x,_rulerdiff+8);
    Canvas.LineTo(fRuler_x+x-3,_rulerdiff+2);
    Canvas.FloodFill(fRuler_x+x,_rulerdiff+5,clRed,fsBorder);

    //Bottom
    Canvas.Moveto(fRuler_x+x-3,2*fruler_y+fImage.Height-2-_rulerdiff);
    Canvas.LineTo(fRuler_x+x+3,2*fruler_y+fImage.Height-2-_rulerdiff);
    Canvas.LineTo(fRuler_x+x,2*fruler_y+fImage.Height-8-_rulerdiff);
    Canvas.LineTo(fRuler_x+x-3,2*fruler_y+fImage.Height-2-_rulerdiff);
    Canvas.FloodFill(fRuler_x+x,2*fruler_y+fImage.Height-_rulerdiff-5,clRed,fsBorder);
    //Left
    Canvas.Moveto(_rulerdiff+2,fRuler_y+y-3);
    Canvas.LineTo(_rulerdiff+2,fRuler_y+y+3);
    Canvas.LineTo(_rulerdiff+8,fRuler_y+y);
    Canvas.LineTo(_rulerdiff+2,fRuler_y+y-3);
    Canvas.FloodFill(_rulerdiff+5,fruler_y+y,clRed,fsBorder);
    //Right
    Canvas.Moveto(2*fruler_x+fImage.Width-2-_rulerdiff,fRuler_y+y-3);
    Canvas.LineTo(2*fruler_x+fImage.Width-2-_rulerdiff,fRuler_y+y+3);
    Canvas.LineTo(2*fruler_x+fImage.Width-8-_rulerdiff,fRuler_y+y);
    Canvas.LineTo(2*fruler_x+fImage.Width-2-_rulerdiff,fRuler_y+y-3);
    Canvas.FloodFill(2*fruler_x+fImage.Width-_rulerdiff-5,fruler_y+y,clRed,fsBorder);

    _DrawSelection(fOrigin, foMovePt);

    if fShowRulerDots then
    begin
    fImage.Canvas.Pen.Style:=psDot;
    fImage.Canvas.Pen.Mode:=pmNot;

    fImage.Canvas.MoveTo(0,y);
    fImage.Canvas.LineTo(fImage.Width,y);
    fImage.Canvas.MoveTo(x,0);
    fImage.Canvas.LineTo(x,fImage.Height);
    end;
    _rdelta_x:=x+fRuler_x;
    _rdelta_y:=y+fruler_y;
    Canvas.Pen:=tp;
    fImage.Canvas.Pen:=tp2;

end;

procedure TExtImage._ImageMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState;X, Y: Integer);
//---------------------------------------------------------------
//  Purpose: Mouse up event
//---------------------------------------------------------------
begin
 if button=mbLeft then
 begin
  fShift:=Shift;

  if fZoom>100 then fOriginalX:=fx+Round((X / fZFactor)) else fOriginalX
  :=Round(fx+(X/fZFactor));
  if fZoom>100 then fOriginalY:=fy+Round((Y / fZFactor)) else fOriginalY
  :=Round(fy+(Y/fZFactor));

  fRelativeX:=X;
  fRelativeY:=Y;
  fCursorX:=fOriginalx;
  fCursorY:=fOriginaly;

  fzmDestPt:=Point(X,Y);
  fDestPt.X:=fOriginalX;
  fDestPt.Y:=fOriginalY;

  if fSelectMode and fDrawSelection then
  begin
    if fSelected then fTimer.Enabled:=True;
    if Assigned(fOnAfterSelect) then fonAfterSelect(Self);
  end;
  fDrawSelection:=False;
  fDrawing:=False;
 end;
  if Assigned(fOnImageMouseUp) then fOnImageMouseUp(self,button,shift,x,y);
end;

procedure TExtImage._ImageMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState;X, Y: Integer);
//---------------------------------------------------------------
//  Purpose: Mouse down event
//---------------------------------------------------------------
begin
 if button=mbLeft then
 begin
  fTimer.Enabled:=False;
  fShift:=Shift;

  if fZoom>100 then fOriginalX:=fx+Round((X / fZFactor)) else fOriginalX
  :=Round(fx+(X/fZFactor));
  if fZoom>100 then fOriginalY:=fy+Round((Y / fZFactor)) else fOriginalY
  :=Round(fy+(Y/fZFactor));
  fCursorX:=fOriginalx;
  fCursorY:=fOriginaly;

  fRelativeX:=X;
  fRelativeY:=Y;
  fDrawSelection:=false;
  if fSelectMode then
  begin
    if fSelected then _ClearSelection((Zoom/100));
    fZoomEventOccured:=False;
    if (ssLeft in Shift) then fDrawSelection:=true;
  end;
  fOrigin.X:=fOriginalX;
  fOrigin.Y:=fOriginalY;
  fzmOrigin := Point(X,Y);
  fzmDestPt := Point(X,Y);
  fOMovePt:=fOrigin;
  fMovePt:=fZMOrigin;
  fDrawing:=True;
 end;
  if Assigned(fOnImageMouseDown) then fOnImageMouseDown(self,button,shift,x,y);
end;

procedure TExtImage._ImageMouseClick(Sender: TObject);
//---------------------------------------------------------------
//  Purpose: Mouse click event
//---------------------------------------------------------------
begin
  if Assigned(fOnImageClick) then fOnImageClick(self);
end;

procedure TExtImage.UpdatePalette;
//---------------------------------------------------------------
//  Purpose: Update palette (Currently only for testing)
//---------------------------------------------------------------
begin
{ if GetDeviceCaps(GetDC(0),BITSPIXEL)=8 then
  begin
    Image.Picture.Bitmap.PixelFormat:=pf8Bit;
    Image.Picture.Bitmap.Palette:=
    LoadOptimizedPalette;
    Image.Picture.Bitmap.IgnorePalette:=false;
    ApplyPaletteToDC(GetDC(self.Handle),Image.Picture.Bitmap.Palette);
  end;}
end;

procedure TExtImage.GetDimensions(AOrigin,ADestPt:TPoint;var rWidth,rHeight:Integer;var rRect:TRect);
begin
  if AOrigin.X> ADestPt.X then begin
     rRect.Right:=AOrigin.X;
     rRect.Left:=ADestPt.X;
  end else
  begin
     rRect.Left:=AOrigin.X;
     rRect.Right:=ADestPt.X;
  end;

  if AOrigin.Y> ADestPt.Y then begin
     rRect.Bottom:=AOrigin.Y;
     rRect.Top:=ADestPt.Y;
  end else
  begin
     rRect.Top:=AOrigin.Y;
     rRect.Bottom:=ADestPt.Y;
  end;
  rWidth:=rRect.Right-rRect.Left;
  rHeight:=rRect.Bottom-rRect.Top;
end;

procedure TExtImage.ScaleRect(var rRect:TRect);
begin
 with rRect do
 begin
    Left:= Round((Left-fx)*fzFactor);
    Right:=Round((Right-fx)*fzFactor) ;
    Top:=Round((Top-fy)*fzFactor);
    Bottom:=Round((Bottom-fy)*fzFactor);
 end;
end;

procedure TExtImage._ClearSelection(Factor:Real);
begin
  if not fblink_active then exit;
  fblink_active:=false;
  if fSelected then
  begin
    fTimer.Enabled:=False;
    fSelected:=False;
    Redraw;
    _ClearRulerDots;
   end;
end;

procedure TExtImage._MakeSelection(Factor:Real;Mode:TPenMode);
begin
   _DrawSelection(fOrigin, fDestPt);
end;

procedure TExtImage._DrawSelection(TopLeft, BottomRight: TPoint);
var rrect:TRect;
    w,h:Integer;
begin
   if (not fSelectMode and not fSelected) or not fSelected or fblink_active then exit;
   fimage.picture.Bitmap.Canvas.CopyMode:=cmSrcCopy;

   GetDimensions(TopLeft,BottomRight,w,h,rrect);
   ScaleRect(rrect);
    with rrect do
    begin
      if Left<0 then Left:=0;
      if Top<0 then Top:=0;
      if Right>fImage.Width then Right:=fImage.Width;
      if Bottom>fImage.Height then Bottom:=fImage.Height;
    end;

    fImage.Canvas.Pen.Style:=psSolid;
    fImage.Canvas.Pen.Color:=clWhite;
    fImage.Canvas.Pen.Mode:=pmNot;

    fImage.Canvas.MoveTo(RasterX(rrect.Left),RasterY(rrect.Top));
    fImage.Canvas.LineTo(RasterX(rrect.right),RasterY(rrect.Top));

    fImage.Canvas.MoveTo(RasterX(rrect.Left),RasterY(rrect.bottom));
    fImage.Canvas.LineTo(RasterX(rrect.right),RasterY(rrect.bottom));

    fImage.Canvas.MoveTo(RasterX(rrect.Left),RasterY(rrect.Top));
    fImage.Canvas.LineTo(RasterX(rrect.Left),RasterY(rrect.bottom));

    fImage.Canvas.MoveTo(RasterX(rrect.right),RasterY(rrect.Top));
    fImage.Canvas.LineTo(RasterX(rrect.right),RasterY(rrect.bottom));
end;

procedure TExtImage._BlinkSelection;
var i,w,h,s,l,d,m:Integer;
    f,d1,d2:Boolean;
    rrect:Trect;
procedure Buffer;
begin

    fblink_active:=true;
end;
begin
   if not fSelected and not fSelectMode then exit;
    GetDimensions(fOrigin,fDestPt,w,h,rrect);
   if not fblink_active then Buffer;
      fImage.Canvas.Pen.Mode:=TPenMode(4);
      fImage.Canvas.Pen.Width:=1;
      fImage.Canvas.Pen.Style:=psSolid;


    fff:=not fff;


   ScaleRect(rrect);
   with rrect do
   begin
     if Left<0 then Left:=0;
     if Top<0 then Top:=0;
     if Right>fImage.Width then Right:=fImage.Width;
     if Bottom>fImage.Height then Bottom:=fImage.Height;
   end;

   sel_point_tl:=rrect.topleft ;
   sel_point_br:=rrect.bottomright;

    //calculate bounds for horizontal lines
    if RasterX(sel_point_tl.X)< RasterX(sel_point_br.X) then
    begin
      m:=1;
      l:=RasterX(sel_point_br.X)-RasterX(sel_point_tl.X);
      if (fOrigin.Y < fy) then d1:=false else d1:=true;
      if (fDestPt.Y > fy + fVisibleVert) then d2:=false else d2:=true;
    end else
    begin
      m:=-1;
      l:=RasterX(sel_point_tl.X)-RasterX(sel_point_br.X);
      if (fOrigin.Y < fy) then d1:=false else d1:=true;
      if (fDestPt.Y > fy + fVisibleVert) then d2:=false else d2:=true;
    end;


    //draw horizontal lines
    f:=fff;
    d:=3;
    s:=1;
    for i:=s to l do
    begin
      if i mod Abs(d) = 0 then
      begin
        if f then fImage.Canvas.Pen.Color:=clBlack else fImage.Canvas.Pen.Color:=clWhite;
        if i+d>l then d:=l-i;
        if d1 then
        begin
          fImage.Canvas.MoveTo(RasterX(sel_point_tl.X)+m*i,RasterY(sel_point_tl.Y));
          fImage.Canvas.LineTo(RasterX(sel_point_tl.X)+m*(i+d),RasterY(sel_point_tl.Y));
        end;
        if d2 then
        begin
          fImage.Canvas.MoveTo(RasterX(sel_point_tl.X)+m*i,RasterY(sel_point_br.Y));
          fImage.Canvas.LineTo(RasterX(sel_point_tl.X)+m*(i+d),RasterY(sel_point_br.Y));
        end;
        f:=not f;
        if d=0 then break;
      end;
    end;

    //calculate bounds for vertical lines
    s:=1;
    if RasterY(sel_point_tl.Y)< RasterY(sel_point_br.Y) then
    begin
      m:=1;
      CheckRasterBoundsY(sel_point_tl.y,sel_point_br.y,0,fImage.Height);
      l:=RasterY(sel_point_br.Y)-RasterY(sel_point_tl.Y);
      if (fOrigin.X < fx) then d1:=false else d1:=true;
      if (fDestPt.X > fx +fVisibleHorz ) then d2:=false else d2:=true;
    end else
    begin
      m:=-1;
      CheckRasterBoundsY(sel_point_tl.y,sel_point_br.y,0,fImage.Height);
      l:=RasterY(sel_point_tl.Y)-RasterY(sel_point_br.Y);
      if (fOrigin.X < fx) then d1:=false else d1:=true;
      if (fDestPt.X > fx +fVisibleHorz ) then d2:=false else d2:=true;
    end;

    //draw vertical lines
    f:=fff;
    d:=3;
    for i:=s to l do
    begin
      if i mod Abs(d) = 0 then
      begin
        if f then fImage.Canvas.Pen.Color:=clBlack else fImage.Canvas.Pen.Color:=clWhite;
        if i+d>l then d:=l-i;
        if d1 then
        begin
          fImage.Canvas.MoveTo(RasterX(sel_point_tl.X),RasterY(sel_point_tl.Y)+m*i);
          fImage.Canvas.LineTo(RasterX(sel_point_tl.X),RasterY(sel_point_tl.Y)+m*(i+d));
        end;
        if d2 then
        begin
          fImage.Canvas.MoveTo(RasterX(sel_point_br.X),RasterY(sel_point_tl.Y)+m*i);
          fImage.Canvas.LineTo(RasterX(sel_point_br.X),RasterY(sel_point_tl.Y)+m*(i+d));
        end;
        f:=not f;
        if d=0 then break;
      end;
    end;
end;

procedure TExtImage.UpdateSelection;
begin
 fSelected:=true;
 _MakeSelection(fZFactor,pmXOR);
end;

procedure TExtImage._BitmapChanged(Sender:TObject);
begin
  if Assigned(FOnBitmapChanged) then FOnBitmapChanged(Self);
end;

procedure TExtImage._SetMonitorGamma(Value:Boolean);
var i:Integer;
begin
  if Value<>fApplyMonitorGamma then
  begin
     fApplyMonitorGamma:=Value;
     for i:=1 to 255 do
     fGammaArray[i]:=IntGamma(fMonitorGamma,i);
  end;
end;

procedure TExtImage._SetMonitorGammaVal(Value:Double);
var i:Integer;
begin
  if Value<>fMonitorGamma then
  begin
     fMonitorGamma:=Value;
     for i:=1 to 255 do
     fGammaArray[i]:=IntGamma(fMonitorGamma,i);
  end;
end;

function TExtImage.RasterX(Val:Integer):Integer;
var d:Integer;
begin
  if FZoom>100 then
  begin
      d:=fPixelWidth;
    Result := Round((Val div d) * d);
    if Result>=fImage.Width then Result:=fImage.Width-1;
  end
  else Result:=Val;

end;

function TExtImage.RasterY(Val:Integer):Integer;
var d:Integer;
begin
  if FZoom>100 then
  begin
    d:=fPixelWidth;
    Result := Round((Val div d) * d);
  if Result>=fImage.Height then Result:=fImage.Height-1;
  end
  else Result:=Val;
end;

{procedure TExtImage.CheckRasterBoundsX(var l1,l2:Integer;Min,Max:Integer);
begin
  exit;
  if RasterX(l1) < Min then
  begin
    l1:=Min;
    if RasterX(l1) < Min then
    l1 := l1 + fHorzPixel;
  end;

  if RasterX(l2) > Max then begin
    l2:=Max;
    if RasterX(l2) > Max then
    l2 := l2 - fHorzPixel;
  end;
end;
 }
procedure TExtImage.CheckRasterBoundsY(var l1,l2:Integer;Min,Max:Integer);
begin
exit;
  if RasterY(l1) < Min then
  begin
    l1:=Min;
    if RasterY(l1) < Min then
    l1 := l1 + fVertPixel;
  end;

  if RasterY(l2) > Max then begin
    l2:=Max;
    if RasterY(l2) > Max then
    l2 := l2 - fVertPixel;
  end;
end;


procedure TExtImage._OnTimer(Sender: TObject);
begin
  _BlinkSelection;
end;

{$ifdef USE_FULL_CAPABILITY}
procedure TExtImage.ReduceColors(ColorReduction:TColorReduction;DitherMode:TDitherMode;ReductionBits:Integer;CustomPalette:HPalette);
var TempBitmap:TBitmap;
begin
  TempBitmap:=TBitmap.Create;
  try
    gfx_colors.ReduceColors(fBitmap,TempBitmap,ColorReduction,DitherMode,ReductionBits,CustomPalette);
    fBitmap.Assign(TempBitmap);
  finally
    TempBitmap.Free;
  end;
end;
{$endif}

procedure Register;
//---------------------------------------------------------------
//  Purpose: Register
//---------------------------------------------------------------
begin
  RegisterComponents('AM FX Pro', [TExtImage]);
end;

end.
