{***********************************************************************

Unit gfx_Effects.PAS v1.5 0801
    (c) by Andreas Moser, amoser@amoser.de,

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_Effects is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    for more information about image processing look at this book:

    Computer Graphics - Principles and Practice, Foley et al

********************************************************************************}

UNIT gfx_effects;

INTERFACE

USES gfx_basedef,
     gfx_syswrapper,
     gfx_dibProcessing;

    procedure EffectAddNoise(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;MonoNoise:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectAntiAlias(SrcBitmap,DestBitmap:TXBitmap;R,G,B:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectArithmetic(SrcBitmapOne,SrcBitmapTwo,DestBitmap:TXBitmap;Arithmetic:TGraphicArithmetic;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectBlend(SrcBitmapOne,SrcBitmapTwo,DestBitmap:TXBitmap;Multiplicator,MaxBlendValue:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectCircleAround(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectColorize(SrcBitmap,DestBitmap:TXBitmap;Hue,Saturation,Lightness:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectContrast(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;R,G,B:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectEllipse(SrcBitmap,DestBitmap:TXBitmap;Radius:Double;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectFishEye(SrcBitmap,DestBitmap:TXBitmap;Value:Extended;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectGamma(SrcBitmap,DestBitmap:TXBitmap;Gamma:Double;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectGreyScale(SrcBitmap,DestBitmap:TXBitmap;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectOldLook(SrcBitmap,DestBitmap:TXBitmap;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectIncDecRGB(SrcBitmap,DestBitmap:TXBitmap;dR,dG,dB:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectLightness(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;R,G,B:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectMosaic(SrcBitmap,DestBitmap:TXBitmap;Width,Height:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectNegative(SrcBitmap,DestBitmap:TXBitmap;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectPosterize(SrcBitmap,DestBitmap:TXBitmap;BitsPerChannel:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectSinus(SrcBitmap,DestBitmap:TXBitmap;SinusAmpVert,VertDelta,SinusAmpHorz,HorzDelta,VertStart,HorzStart:Integer;ChngVertAtAnyCol:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectSolarize(SrcBitmap,DestBitmap:TXBitmap;Threshold:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectSpray(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectStretch(SrcBitmap,DestBitmap:TXBitmap;Low,High:Integer;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectSwapChannels(SrcBitmap,DestBitmap:TXBitmap;WhichPlanes:TSwapPlanes;const EffectCallBack:TEffectCallBack);stdcall;

    {Filter functions: For universal processing use EffectFilter..the class looks for itself which filter will be used}
    procedure EffectFilter(SrcBitmap,DestBitmap:TXBitmap;Filter:TGraphicFilter;Size:TMatrixSize;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectStaticFilter(SrcBitmap,DestBitmap:TXBitmap;StaticFilterType:TStaticFilterType;Diameter:Integer;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectMultiPass(SrcBitmap,DestBitmap:TXBitmap;MultiPassFilter:TMultiPassGraphicFilter;Passes:Integer;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
    procedure EffectLinearFilter(SrcBitmap,DestBitmap:TXBitmap;Filter:TGraphicFilter;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;

IMPLEMENTATION

// -----------------------------------------------------------------------------
//
//			Blending two Bitmaps
//
//     Parameter:
//               SrcBitmapOne  : 1st Bitmap
//               SrcBitmapOne  : 2nd Bitmap
//               DestBitmap    : Result
//               Multiplicator,
//               MaxBlendValue  : The mixing of the two pictures...
//               EffectCallBack  :  CallBack for user interface
// -----------------------------------------------------------------------------
procedure EffectBlend(SrcBitmapOne,SrcBitmapTwo,DestBitmap:TXBitmap;Multiplicator,MaxBlendValue:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface2;
begin

   if not CheckPixelFormat(SrcBitmapOne,xpf24bit) then Exit;
   if not CheckPixelFormat(SrcBitmapTwo,xpf24bit) then Exit;
    BMPIntfc:=TBMPInterface2.Create(SrcBitmapOne,SrcBitmapTwo,DestBitmap,xpf24Bit,0,True);
    try
      DIBEffectArithmetic(BMPIntfc,gaBlend,true,true,true,Multiplicator,MaxBlendValue,EffectCallBack);
    finally
      BMPIntfc.Free;
    end;

end;

// -----------------------------------------------------------------------------
//
//			Arithmetic between two Bitmaps
//
//     Parameter:
//               SrcBitmapOne  : 1st Bitmap
//               SrcBitmapOne  : 2nd Bitmap
//               DestBitmap    : Result
//               Arithmetic    : Arithmetic type
//               EffectCallBack  :  CallBack for user interface
// -----------------------------------------------------------------------------
procedure EffectArithmetic(SrcBitmapOne,SrcBitmapTwo,DestBitmap:TXBitmap;Arithmetic:TGraphicArithmetic;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface2;
begin
   if not CheckPixelFormat(SrcBitmapOne,xpf24bit) then Exit;
   if not CheckPixelFormat(SrcBitmapTwo,xpf24bit) then Exit;
    BMPIntfc:=TBMPInterface2.Create(SrcBitmapOne,SrcBitmapTwo,DestBitmap,xpf24Bit,0,True);
    try
      DIBEffectArithmetic(BMPIntfc,Arithmetic,true,true,true,0,0,EffectCallBack);
    finally
      BMPIntfc.Free;
    end;
end;

// -----------------------------------------------------------------------------
//
//			Greyscale Bitmap
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectGreyScale(SrcBitmap,DestBitmap:TXBitmap;const EffectCallBack:TEffectCallBack);stdcall;
var
 BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxGrayScale,True,True,True,0,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			OldLook
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectOldLook(SrcBitmap,DestBitmap:TXBitmap;const EffectCallBack:TEffectCallBack);stdcall;
var
 BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,SrcBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc, fxOldLook,True,True,True,0,false,EffectCallBack);
  finally
    DestBitmap.Assign(SrcBitmap);
    BMPIntfc.Free;

  end;
end;

// -----------------------------------------------------------------------------
//
//			Posterize Bitmap
//
//     Parameter:
//               SrcBitmap      : Bitmap to be processed
//               DestBitmap     : Result
//               BitsPerChannel : How many Bits per channel
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectPosterize(SrcBitmap,DestBitmap:TXBitmap;BitsPerChannel:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
    BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxPosterize,True,True,True,BitsPerChannel,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Solarize Bitmap
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Threshold    : Process only pixels above this value
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectSolarize(SrcBitmap,DestBitmap:TXBitmap;Threshold:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
 BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxSolarize,True,True,True,Threshold,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Negative Bitmap
//
//     Parameter:  (Remark: this is just the same as Solarize, just a threshold of zero)
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectNegative(SrcBitmap,DestBitmap:TXBitmap;const EffectCallBack:TEffectCallBack);stdcall;
var
 BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxSolarize,True,True,True,0,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Lighten/Darken Bitmap
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Value        : Value to be added/substracted (-255..0..255)
//               R,G,B        : Apply filter to R,G,B Planes (Boolean)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectLightness(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;R,G,B:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxLightness,r,g,b,Value,false,EffectCallBack);
  finally
    BmpIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Change Contrast
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Value        : Value to be added/substracted (-255..0..255)
//               R,G,B        : Apply filter to R,G,B Planes (Boolean)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectContrast(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;R,G,B:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxContrast,r,g,b,Value,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Colorize Bitmap
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Hue          : Value to be added to Hue (0..359)
//               Saturation   : Value to be added to Saturation (0..255)
//               Value        : Value to be added to Value (0..255)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectColorize(SrcBitmap,DestBitmap:TXBitmap;Hue,Saturation,Lightness:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;

begin
    if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect3Val(BMPintfc,fxColorize,true,true,true,Hue,Saturation,Lightness,false,EffectCallBack);
  finally
    BmpIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			IncDecRGB Bitmap
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               dR           : Value to be added to Red (0..255)
//               dG           : Value to be added to Green (0..255)
//               dB           : Value to be added to Blue (0..255)
//               EffectCallBack  :  CallBack for user interface
//
//     (c) September 1998 by Andreas Moser, amoser@amoser.de
// -----------------------------------------------------------------------------
procedure EffectIncDecRGB(SrcBitmap,DestBitmap:TXBitmap;dR,dG,dB:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect3Val(BMPintfc,fxIncdecRGB,true,true,true,dR,dG,dB,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Apply filter to Bitmap
//                      (just an interface for the 3 following procedures)
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Size         : Optional:Size of the calculated area 1..7 (neighbours of the pixel)
//                              if set to 0, the default size for the filter will be used
//                              (normally for linear 3x3, for static filters a diameter of 3)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectFilter(SrcBitmap,DestBitmap:TXBitmap;Filter:TGraphicFilter;Size:TMatrixSize;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  d                       :Integer;
  sft                     :TStaticFilterType;
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try

    case Filter.FilterType of
    ftLinear: DIBEffectLinearFilter(BmpIntfc,Filter,ColorSpace,Channel1,Channel2,Channel3,channel4,EffectCallBack);
    ftStatic:
      begin
       d:=3;
       CASE Size of
          mx3:d:=3;
          mx5:d:=5;
          mx7:d:=7;
        end;
        if Filter.FilterName='Minimum (Static)' then sft:=sfMin ELSE if
        Filter.FilterName='Maximum (Static)' then sft:=sfMax ELSE sft:=sfMedian;
        DIBEffectStaticFilter(BmpIntfc,sft,d,ColorSpace,Channel1,Channel2,Channel3,Channel4,EffectCallBack);
      end;
    end;
  finally
    BmpIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Apply Median static filter to Bitmap
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Diameter     : Size of the calculated area 1..7 (neighbours of the pixel)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectStaticFilter(SrcBitmap,DestBitmap:TXBitmap;StaticFilterType:TStaticFilterType;Diameter:Integer;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
    DIBEffectStaticFilter(BmpIntfc,StaticFilterType,Diameter,ColorSpace,Channel1,Channel2,Channel3,Channel4,EffectCallBack);
  finally
    BmpIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Apply linear filter to Bitmap (RGB,HSV,HSL)
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Filter       : Filter (see TGraphicFilter at the top of this unit)
//               ColorSpace   : Which colorspase should be used (csRGB,csHSV,csHSL)
//               Channel 1-3  : Apply filter to Channel 1 -3
//                              csRGB R:Channel1 G:Channel 2 B: Channel 3
//                              csHSV H:Channel1 S:Channel 2 V: Channel 3
//                              csHSL H:Channel1 S:Channel 2 L: Channel 3
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectLinearFilter(SrcBitmap,DestBitmap:TXBitmap;Filter:TGraphicFilter;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
   if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
   if (@Filter=nil) OR (Filter.Divisor=0) OR (Filter.FilterType<>ftLinear) then
   begin
     DestBitmap.Assign(SrcBitmap);
     Exit;
   end;
   BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
   try
     DIBEffectLinearFilter(BmpIntfc,Filter,ColorSpace,Channel1,Channel2,Channel3,channel4,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;



// -----------------------------------------------------------------------------
//
//			Apply multi-pass filter to Bitmap (RGB-Model)
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               MultiPassFilter : MultiPassFilter (see TMultiPassGraphicFilter at the top of this unit)
//               R,G,B        : Apply filter to R,G,B Planes (Boolean)
//               Passes       : Nr of filters to be applied
//               EffectCallBack  :  CallBack for user interface
//
//     
// -----------------------------------------------------------------------------
procedure EffectMultiPass(SrcBitmap,DestBitmap:TXBitmap;MultiPassFilter:TMultiPassGraphicFilter;Passes:Integer;ColorSpace:TColorSpace;Channel1,Channel2,Channel3,Channel4:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var  Bmp1,Bmp2,Bmp3:TXBitmap;

procedure ClearBitmap(Bitmap:TXBitmap);
begin
  Bitmap.Canvas.Brush.Color:=xclWhite;
  Bitmap.Canvas.FillRect(xRect(0,0,Bitmap.Width,Bitmap.Height));
end;

begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  if not (passes in [1..4]) OR (MultiPassFilter.FilterType<>ftMultiPass)  then exit;
  Bmp1:=TXBitmap.Create;
  Bmp2:=TXBitmap.Create;
  Bmp3:=TXBitmap.Create;
  Bmp1.PixelFormat:=xpf24bit;
  Bmp2.PixelFormat:=xpf24bit;
  Bmp3.PixelFormat:=xpf24bit;
  Bmp1.Assign(SrcBitmap);
  try
  if Passes>=2 then
  begin
    EffectFilter(Bmp1,Bmp2,MultiPassFilter.Filters[1]^,mxDefault,ColorSpace,Channel1,Channel2,Channel3,Channel4,EffectCallBack);
    EffectFilter(Bmp2,Bmp3,MultiPassFilter.Filters[2]^,mxDefault,ColorSpace,Channel1,Channel2,Channel3,Channel4,EffectCallBack);
    ClearBitmap(Bmp1);
    if MultiPassFilter.Functions[1]<>gaNone then EffectArithmetic(bmp2,bmp3,bmp1,MultiPassFilter.Functions[1],EffectCallBack)
    ELSE Bmp1.Assign(Bmp3);
    ClearBitmap(Bmp2);
    ClearBitmap(Bmp3);
    if Passes>=3 then
    begin
      EffectFilter(Bmp1,Bmp2,MultiPassFilter.Filters[3]^,mxDefault,ColorSpace,Channel1,Channel2,Channel3,Channel4,EffectCallBack);
      if MultiPassFilter.Functions[2]<>gaNone then
      begin
        EffectArithmetic(bmp1,bmp2,bmp3,MultiPassFilter.Functions[2],EffectCallBack);
        ClearBitmap(Bmp1);
        Bmp1.Assign(Bmp3);
      end ELSE
      begin
        ClearBitmap(Bmp1);
        Bmp1.Assign(Bmp2);
      end;
      ClearBitmap(Bmp3);
      ClearBitmap(Bmp2);
      if Passes = 4 then
      begin
        EffectFilter(Bmp1,Bmp2,MultiPassFilter.Filters[4]^,mxDefault,ColorSpace,Channel1,Channel2,Channel3,Channel4,EffectCallBack);
        if MultiPassFilter.Functions[3]<>gaNone then
        begin
          EffectArithmetic(bmp1,bmp2,bmp3,MultiPassFilter.Functions[3],EffectCallBack);
          ClearBitmap(Bmp1);
          Bmp1.Assign(Bmp3);
        end ELSE
        begin
          ClearBitmap(Bmp1);
          Bmp1.Assign(Bmp2);
        end;
          ClearBitmap(Bmp3);
          ClearBitmap(Bmp2);
      end;
    end;
  end;
    DestBitmap.Assign(Bmp1);
  finally
    Bmp1.Free;
    Bmp2.Free;
    Bmp3.Free;
  end;
end;


// -----------------------------------------------------------------------------
//
//			Spray effect
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Value        : Intensity
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------

procedure EffectSpray(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;

begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
   DIBEffect1Val(BMPintfc,fxSpray,true,true,true,Value,false,EffectCallBack);
  finally
   BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Add noise
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               MonoNoise    : Should noisevalue calculated for each plane or not
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------

procedure EffectAddNoise(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;MonoNoise:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
    DIBEffect1Val(BMPIntfc,fxAddNoise,true,true,true,Value,MonoNoise,EffectCallback);
  finally
    BMPIntfc.Free;
  end;
end;

// -----------------------------------------------------------------------------
//
//			Anti-aliasing
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               R,G,B        : Apply antialias to R,G,B Planes (Boolean)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------

procedure EffectAntiAlias(SrcBitmap,DestBitmap:TXBitmap;R,G,B:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
    if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
    BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
    try
      DIBEffect1Val(BMPIntfc,fxAntiAlias,R,G,B,0,false,EffectCallback);
    finally
      BMPIntfc.Free;
    end;
end;


// -----------------------------------------------------------------------------
//
//			Stretch
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------

procedure EffectStretch(SrcBitmap,DestBitmap:TXBitmap;Low,High:Integer;const EffectCallBack:TEffectCallBack);stdcall;
var
  BMPIntfc                :TBMPInterface1;
begin
    if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
    BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
    try
      DIBEffect2Val(BMPintfc,fxStretch,true,true,true,Low,High,false,EffectCallBack);
    finally
      BMPIntfc.Free;
    end;
end;


procedure EffectGamma(SrcBitmap,DestBitmap:TXBitmap;Gamma:Double;const EffectCallBack:TEffectCallBack);stdcall;
// -----------------------------------------------------------------------------
//
//			Gamma correction
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Value        : Gamma value
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
var
  BMPIntfc                :TBMPInterface1;
begin
    if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

    BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
    try
      DIBEffect1Val(BMPintfc,fxGamma,true,true,true,Gamma,false,EffectCallBack);
    finally
      BMPIntfc.Free;
    end;
end;



procedure EffectEllipse(SrcBitmap,DestBitmap:TXBitmap;Radius:Double;const EffectCallBack:TEffectCallBack);stdcall;
// -----------------------------------------------------------------------------
//
//			Ellipse
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
var
  BMPIntfc                :TBMPInterface1;

begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
      if Radius= 0 then  Radius:=1;
      DIBEffect1Val(BMPintfc,fxEllipse,true,true,true,Radius,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;


procedure EffectMosaic(SrcBitmap,DestBitmap:TXBitmap;Width,Height:Integer;const EffectCallBack:TEffectCallBack);stdcall;
// -----------------------------------------------------------------------------
//
//			Mosaic effect
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Width,Height : Dimension of blocks
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  if(Width<1) or (Height<1) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
    DIBEffect2Val(BMPintfc,fxMosaic,true,true,true,Width,Height,false,EffectCallBack);
  finally
    BMPIntfc.free;
  end;
end;

procedure EffectCircleAround(SrcBitmap,DestBitmap:TXBitmap;Value:Integer;const EffectCallBack:TEffectCallBack);stdcall;
// -----------------------------------------------------------------------------
//
//			Picture distortion
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Value        : Amount
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
      DIBEffect1Val(BMPintfc,fxCircleAround,true,true,true,Value,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;

procedure EffectFishEye(SrcBitmap,DestBitmap:TXBitmap;Value:Extended;const EffectCallBack:TEffectCallBack);stdcall;
// -----------------------------------------------------------------------------
//
//			FishEye effect
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Value        : Amount
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------

var
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;
  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
      DIBEffect1Val(BMPintfc,fxFishEye,true,true,true,Value,false,EffectCallBack);
  finally
    BMPIntfc.Free;
  end;
end;



// -----------------------------------------------------------------------------
//
//			Sinus effect
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               SinusAmpVer  : Amplitude vertical
//               VertDelta    : Delta vertical
//               SinusAmpHorz : Amplitude horizontal
//               HorzDleta    : Delta horizontal
//               VertStart    : Vertical start
//               VertStart    : Vertical start
//               R,G,B        : Apply antialias to R,G,B Planes (Boolean)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------

procedure EffectSinus(SrcBitmap,DestBitmap:TXBitmap;SinusAmpVert,VertDelta,SinusAmpHorz,HorzDelta,
                      VertStart,HorzStart:Integer;
                      ChngVertAtAnyCol:Boolean;const EffectCallBack:TEffectCallBack);stdcall;
var
  Row,Col,s,ds,t,dt :Integer;
  TargetRow       :pRGBArray;
  BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
    dt:=HorzStart;

    if VertDelta=0 then VertDelta:=1;
    if HorzDelta=0 then HorzDelta:=1;

    for Row := 0 to BmpIntfc.Height - 1 do
    begin
      if not ChngVertAtAnyCol then if dt/HorzDelta<360 then inc(dt,1) else dt:=0;

      if Assigned(EffectCallBack) then EffectCallBack(0,100,Round((Row/SrcBitmap.Height)*100));

       TargetRow := BMPIntfc.DestScanline[Row];
       ds:=VertStart;
       for Col := 0 to BmpIntfc.Width - 1 do
       begin
       s:=Round(SinusAmpVert*Sin(ds / VertDelta));
       if ds/VertDelta<360 then inc(ds,1) else ds:=0;
       if ChngVertAtAnyCol then if dt/HorzDelta<360 then inc(dt,1) else dt:=0;
       if (Row+s<0) or (Row+s> BmpIntfc.Height - 1 )then s:=0;
       t:=Round(SinusAmpHorz*Sin(dt / HorzDelta));
       if (Col+t<0) or (Col+t> BmpIntfc.Width - 1 )then t:=0;
       TargetRow[Col] := BMPIntfc.SourceRGBTriple[Row+S,Col+T];
    end;
    end;
  finally
    BmpIntfc.Free;
  end;
end;


// -----------------------------------------------------------------------------
//
//			Swap colorchannels
//
//     Parameter:
//               SrcBitmap    : Bitmap to be processed
//               DestBitmap   : Result
//               Which planes : see TSwapPlanes (at the top of this unit)
//               EffectCallBack  :  CallBack for user interface
//
//
// -----------------------------------------------------------------------------
procedure EffectSwapChannels(SrcBitmap,DestBitmap:TXBitmap;WhichPlanes:TSwapPlanes;const EffectCallBack:TEffectCallBack);stdcall;
var Row,Col        :Integer;
    SrcRow,DestRow :pRGBArray;
    BMPIntfc                :TBMPInterface1;
begin
  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24bit,0,True);
  try
  for Row:=0 to BMPIntfc.Height-1 do
  begin

     if Assigned(EffectCallBack) then EffectCallBack(0,100,Round((Row/SrcBitmap.Height)*100));

     SrcRow:=BmpIntfc.SourceScanLine[Row];
     DestRow:=BmpIntfc.DestScanLine[Row];
     for Col:=0 to BMPIntfc.Width-1 do
     begin
       WITH DestRow[Col] do
         CASE WhichPlanes of
           swapRedGreen:begin
               rgbtBlue:=SrcRow[Col].rgbtBlue;
               rgbtGreen:=SrcRow[Col].rgbtRed;
               rgbtRed:=SrcRow[Col].rgbtBlue;
             end;
           swapRedBlue:begin
               rgbtBlue:=SrcRow[Col].rgbtRed;
               rgbtGreen:=SrcRow[Col].rgbtGreen;
               rgbtRed:=SrcRow[Col].rgbtBlue;
             end;
           swapBlueGreen:begin
               rgbtBlue:=SrcRow[Col].rgbtGreen;
               rgbtGreen:=SrcRow[Col].rgbtBlue;
               rgbtRed:=SrcRow[Col].rgbtRed;
             end;
         end;
     end;
  end;
  finally
    bmpIntfc.Free;
  end;
end;

end.

